/* Name: cdc162.c
 * Project: AVR USB driver for AT90USB
 * Author: Koji Suzuki
 * Creation Date: 2007-12-9
 * Tabsize: 8
 * Copyright: (c) 2007 by Koji Suzuki (suz@nmj.sakura.ne.jp)
 * License: Proprietary, free under certain conditions. See Documentation.
 */

// 2009.4.12	KAWAKAMI Yukio
//	usbcdc_setup()にボーレート変更コールバック追加
//	usbcdc_getc()の戻り値を unsigned char に変更
//	usbcdc_putc()の引数を unsigned char に変更
//	usbcdc_init()に wdr() を追加

#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include "usbconfig.h"
#include "usb162.h"
#include "cdc162.h"

// ボーレート変更コールバック
void callback_set_baud(long baud);


#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))


#define nop()	__asm__ volatile ("nop" )
#define	wdr()	__asm__ volatile("wdr" )

uint8_t usbcdc_modebuf[7];
#define CDCMODE_BAUD0		0
#define CDCMODE_BAUD1		1
#define CDCMODE_BAUD2		2
#define CDCMODE_BAUD3		3
#define CDCMODE_STOPBIT		4
#define CDCMODE_PARITY		5
#define CDCMODE_DATABIT		6

static int8_t usbcdc_setup(uint8_t *in, uint8_t *out){
	if ( (in[USB_RQ_TYPE] & USB_RQ_TYPE_MASK) == USB_RQ_TYPE_CLASS) {
	    if (in[USB_RQ_REQ] == USB_RQ_COMM_GET_LINE) {
			memcpy((void *)out, (const void *)usbcdc_modebuf, 7);
			return 7;
	    } else if (in[USB_RQ_REQ] == USB_RQ_COMM_SET_LINE) {
			int8_t i;
			while (bit_is_clear(UEINTX, RXOUTI))
				;

			for (i=0; i<7; i++) {
				usbcdc_modebuf[i] = UEDATX;
			}

			bit_clear(UEINTX, RXOUTI);

			// ボーレート設定コールバック
			callback_set_baud(*((long *)&usbcdc_modebuf[0]));

			return -1;  // send zlp
	    } else if (in[USB_RQ_REQ] == USB_RQ_COMM_SET_CONTROL) {
			bit_set(GPIOR2, USB162_RX_SEND_EMPTY);
			return -1;  // send zlp only
	    } else {
			return -2; // reply STALL
	    }
	} else {
		return -2; // reply STALL
	}
}

#ifdef EEPROM_LOG
uint16_t log_seq;
#endif

void usbcdc_init() {
#ifdef EEPROM_LOG
    {
	    int16_t i;
	    uint8_t buf[4];
	    memset(buf, 0xff, 4);
	    for (i=0x80; i< 0x100; i+=4) {
		    eeprom_write_block(buf, (void *)i, 4);
	    }
    }
#endif
	usb_user_setup = usbcdc_setup;
	usbInit();
	sei();
	while (!usb_connected) {
		wdr();					// 追加
		usbPoll();
	}
}

void usbcdc_poll() {
	
	UENUM = EP_TX;
	if (bit_is_set(UEINTX,RXOUTI)) {
		bit_clear(UEINTX, RXOUTI);
	}
	if (bit_is_set(UEINTX, FIFOCON) && bit_is_clear(UEINTX,RWAL)) {
		bit_clear(UEINTX, FIFOCON);
	}
	UENUM = EP_RX;
	if (bit_is_set(UEINTX,TXINI)) {
		bit_clear(UEINTX,TXINI);
		bit_clear(UEINTX,NAKINI);
	}
	if (bit_is_set(UDINT, SOFI)) {
		if (UEBCLX) bit_set(GPIOR2, USB162_RX_SEND_REQ);
	}
	if (bit_is_set(UEINTX, FIFOCON)
		&& (bit_is_set(GPIOR2, USB162_RX_SEND_EMPTY)
		    || bit_is_clear(UEINTX,RWAL)
		    || bit_is_set(GPIOR2,USB162_RX_SEND_REQ))) {
#ifdef EEPROM_LOG
		if (log_seq < 0x80) {
			uint8_t data;
			data = (UESTA1X << 6) | (UEBCLX & 0x3f);
			eeprom_write_byte((uint8_t *)(0x80 + log_seq), data);
			log_seq ++;
		}
#endif
		bit_clear(GPIOR2, USB162_RX_SEND_REQ);
		bit_clear(GPIOR2, USB162_RX_SEND_EMPTY);
		if (UEBCLX) bit_set(GPIOR2, USB162_RX_SEND_EMPTY);
		bit_clear(UEINTX, FIFOCON);
	}

	usbPoll();
}

int8_t usbcdc_can_getc(int8_t size) {
	UENUM = EP_TX;
	return (UEBCLX >= size);
}

int8_t usbcdc_can_putc(int8_t size) {
	UENUM = EP_RX;
	return (UEBCLX + size) <= EP_RX_SIZE;
}

unsigned char usbcdc_getc(void) {
	unsigned char ret;
	UENUM = EP_TX;
	while (bit_is_clear(UEINTX,RWAL)) {
		usbcdc_poll();
		UENUM = EP_TX;
#ifdef EEPROM_LOG
		if (bit_is_set(UEINTX,RWAL) && (log_seq < 0x80)) {
			uint8_t data;
			data = (UESTA1X << 6) | 0x80 | (UEBCLX & 0x3f);
			eeprom_write_byte((uint8_t *)(0x80 + log_seq), data);
			log_seq ++;
		}
#endif
	}
	ret = UEDATX;
	return ret;
}

void usbcdc_putc(unsigned char data) {
	UENUM = EP_RX;
	while (bit_is_clear(UEINTX,RWAL)) {
		usbcdc_poll();
		UENUM = EP_RX;
	}
	UEDATX = data;
}
