/* Name: desc.c
 * Project: AVR USB driver for AT90USB
 * Author: Koji Suzuki
 * Creation Date: 2007-12-10
 * Tabsize: 8
 * Copyright: (c) 2006 by Recursion Co., Ltd.
 * Copyright: (c) 2007 by Koji Suzuki (suz@nmj.sakura.ne.jp)
 * Original Project: AVR USB driver for CDC interface on Low-Speed USB
 * Original Author: Osamu Tamura
 * License: Proprietary, free under certain conditions. See Documentation.
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <util/delay.h>

#include "usbconfig.h"
#include "desc.h"

#define USB_LOWSPEED	0

/* -------------------------- Device Description --------------------------- */
#define USB_CFG_EXTERNAL_CONFIG_DESCRIPTOR_LENGH    67
/* Define this to the size of your configuration descriptor if you provide
 * your own configuration descriptor at the global program memory address
 * 'usbDescrConfig'.
 */

#define  USB_CFG_VENDOR_ID       0xc0, 0x16		/* Atmel */
#define  USB_CFG_DEVICE_ID       0xe1, 0x05
#define  USB_CFG_DEVICE_VERSION  0x00, 0x01
#define  USB_CFG_IS_SELF_POWERED 0

#define USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH    0

PROGMEM char usbDescrDevice[] = {    /* USB device descriptor */
    18,         /* sizeof(usbDescrDevice): length of descriptor in bytes */
    USB_DESCR_DEVICE,    /* descriptor type */
    0x01, 0x01, /* USB version supported */
    2,		/* device class */
    0,		/* subclass */
    0,          /* protocol */
#if USB_LOWSPEED
    8,          /* max packet size */
#else
    EP_SIZE,       /* max packet size */
#endif
    USB_CFG_VENDOR_ID,  /* 2 bytes */
    USB_CFG_DEVICE_ID,  /* 2 bytes */
    USB_CFG_DEVICE_VERSION, /* 2 bytes */
    1,          /* manufacturer string index */
    2,          /* product string index */
    0,          /* serial number string index */
    1,          /* number of configurations */
};

PROGMEM char usbDescrConfig[] = {    /* USB configuration descriptor */
    9,          /* sizeof(usbDescrConfig): length of descriptor in bytes */
    USB_DESCR_CONFIG,    /* descriptor type */
    USB_CFG_EXTERNAL_CONFIG_DESCRIPTOR_LENGH,
    0,          /* total length of data returned (including inlined descriptors) */
    2,          /* number of interfaces in this configuration */
    1,          /* index of this configuration */
    0,          /* configuration name string index */
#if USB_CFG_IS_SELF_POWERED
    USB_ATTR_SELFPOWER,  /* attributes */
#else
    USB_ATTR_BUSPOWER,   /* attributes */
#endif
    100/2,            /* max USB current in 2mA units */

    /* interface descriptor follows inline: */
    9,          /* sizeof(usbDescrInterface): length of descriptor in bytes */
    USB_DESCR_INTERFACE, /* descriptor type */
    0,          /* index of this interface */
    0,          /* alternate setting for this interface */
    1,   /* endpoints excl 0: number of endpoint descriptors to follow */
    2, /* CDC Class */
    2, /* ACM (Abstract Modem) */
    1, /* AT-Commands */
    0,          /* string index for interface */

    /* CDC Class-Specific descriptor */
    5,           /* sizeof(usbDescrCDC_HeaderFn): length of descriptor in bytes */
    0x24,        /* descriptor type */
    0,           /* header functional descriptor */
    0x10, 0x01,

    4,           /* sizeof(usbDescrCDC_AcmFn): length of descriptor in bytes */
    0x24,        /* descriptor type */
    2,           /* abstract control management functional descriptor */
    0x02,        /* SET_LINE_CODING,    GET_LINE_CODING, SET_CONTROL_LINE_STATE    */

    5,           /* sizeof(usbDescrCDC_UnionFn): length of descriptor in bytes */
    0x24,        /* descriptor type */
    6,           /* union functional descriptor */
    0,           /* CDC_COMM_INTF_ID */
    1,           /* CDC_DATA_INTF_ID */

    5,           /* sizeof(usbDescrCDC_CallMgtFn): length of descriptor in bytes */
    0x24,        /* descriptor type */
    1,           /* call management functional descriptor */
    3,           /* allow management on data interface, handles call management by itself */
    1,           /* CDC_DATA_INTF_ID */

    /* Endpoint Descriptor */
    7,           /* sizeof(usbDescrEndpoint) */
    USB_DESCR_ENDPOINT,  /* descriptor type = endpoint */
    0x80 | EP_INT,      /* IN endpoint number 3 */
    0x03,        /* attrib: Interrupt endpoint */
#if USB_LOWSPEED
    8, 0,        /* maximum packet size */
#else
    EP_INT_SIZE, 0,        /* maximum packet size */
#endif
    100,         /* in ms */

    /* Interface Descriptor  */
    9,           /* sizeof(usbDescrInterface): length of descriptor in bytes */
    4,           /* descriptor type */
    1,           /* index of this interface */
    0,           /* alternate setting for this interface */
    2,           /* endpoints excl 0: number of endpoint descriptors to follow */
    0x0A,        /* Data Interface Class Codes */
    0,
    0,           /* Data Interface Class Protocol Codes */
    0,           /* string index for interface */

    /* Endpoint Descriptor */
    7,           /* sizeof(usbDescrEndpoint) */
    USB_DESCR_ENDPOINT,  /* descriptor type = endpoint */
    0x80 | EP_RX,        /* IN endpoint number 1 */
    0x02,        /* attrib: Bulk endpoint */
#if USB_LOWSPEED
    8, 0,        /* maximum packet size */
#else
    EP_RX_SIZE, 0,        /* maximum packet size */
#endif
    0,           /* in ms */

    /* Endpoint Descriptor */
    7,           /* sizeof(usbDescrEndpoint) */
    USB_DESCR_ENDPOINT,  /* descriptor type = endpoint */
    EP_TX,       /* OUT endpoint number 2 */
    0x02,        /* attrib: Bulk endpoint */
#if USB_LOWSPEED
  #if UART_CFG_HAVE_USART
    6, 0,        /* maximum packet size 8->6 */
  #else
    1, 0,        /* maximum packet size */
  #endif
#else
    EP_TX_SIZE, 0,        /* maximum packet size */
#endif
    0,           /* in ms */
};

PROGMEM char  usbDescrString0[] = {
	4,		/* sizeof(usbDescrString0) */
	3,		/* descriptor type */
	0x09, 0x04,	/* language index (0x0409 = US-English) */
};

PROGMEM int16_t  usbDescrString1[] = {
    17*2,
    'w', 'w', 'w', '.', 'r', 'e', 'c', 'u', 'r', 's', 'i', 'o', 'n', '.', 'j', 'p' , 0, 
};
PROGMEM int16_t  usbDescrString2[] = {
    8*2,
    'U', 'S', 'B', '-', '2', '3', '2' , 0, 
};


PROGMEM int16_t  usbDescrString3[] = {
	0,
};
