//********************************************************************
//	R8C Series programmer at AT90USB162
//
//	Copyright (c)2009	KAWAKAMI Yukio
//	2009.10.9	Version 0.1.1
//	2009.4.19	Version 0.1.0
//********************************************************************
/*
 * Copyright (c) 2007 Koji Suzuki
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>

#include "usbconfig.h"
#include "cdc162.h"


#define	CPU_CLOCK	F_CPU

// ポート初期設定
#define	PORTB_INIT	0x3E
#define	PORTB_DIR	0xC0
#define	PORTC_INIT	0xFF
#define	PORTC_DIR	0x00
#define	PORTD_INIT	0xFC
#define	PORTD_DIR	0x00

// ポート出力
#define	LED_POWER_ON	PORTB |= (1<<7)
#define	LED_POWER_OFF	PORTB &= ~(1<<7)
#define	LED_PROG_ON		PORTB |= (1<<6)
#define	LED_PROG_OFF	PORTB &= ~(1<<6)
#define	RESET_L			PORTD &= ~(1<<4), DDRD |= (1<<4)
#define	RESET_H			DDRD &= ~(1<<4), PORTD |= (1<<4)
#define	PEN1_L			DDRD |= (1<<0)
#define	PEN1_H			DDRD &= ~(1<<0)
#define	PEN2_L			DDRD |= (1<<1)
#define	PEN2_H			DDRD &= ~(1<<1)
#define	MODE_L			PORTD &= ~(1<<3), DDRD |= (1<<3)
#define	MODE_H			DDRD &= ~(1<<3), PORTD |= (1<<3)
#define	MODE_ON			DDRD |= (1<<5), PORTD &= ~(1<<5)
#define	MODE_OFF		PORTD |= (1<<5), DDRD &= ~(1<<5)
#define	IS_MODE			(DDRD & (1<<5))

// ポート入力
#define	TARGET_POWER	(!(PINB & (1<<5)))		// ターゲット電圧確認

// タイマ設定値
#define	MODE_TIME		250		// max 255(x10ms)
#define	DATA_INTERVAL	100

#define		NOP		asm volatile("nop\n"::)
#define		WDR		asm volatile("wdr\n"::)
#define		SLEEP	asm volatile("sleep\n"::)

// グローバル変数
volatile unsigned char	Timer = 0;				// 汎用タイマ
volatile unsigned char	ModeTimer = 0;			// プログラムモードタイマ
volatile unsigned char	DataIntervalTimer = 0;	// データ間隔タイマ

long	Baud = 0;		// ボーレート設定値

// シリアル受信バッファ
volatile unsigned char	Rbuff[240];		// シリアル受信バッファ
volatile unsigned char	Rremain = 0;	// 受信バッファの残り数
unsigned char			Rin = 0;		// 入力位置
unsigned char			Rout = 0;		// 出力位置

// シリアル送信バッファ
volatile unsigned char	Sbuff[160];
volatile unsigned char	Sremain = 0;
unsigned char			Sin = 0;
unsigned char			Sout = 0;

//********************************************************************
//	100Hz タイマ割り込み
//********************************************************************
ISR(TIMER1_COMPA_vect){

	if (Timer)				--Timer;
	if (ModeTimer)			--ModeTimer;
	if (DataIntervalTimer)	--DataIntervalTimer;
}


//********************************************************************
//	シリアル 受信割り込み
//********************************************************************
ISR(USART1_RX_vect){

	if (UCSR1A & ((1<<FE1)|(1<<DOR1)|(1<<UPE1))){	// 受信エラー
		unsigned char	dummy;
		dummy = UDR1;	// 読み捨て
	} else {
		Rbuff[Rin++] = UDR1;
		if (Rin >= sizeof(Rbuff)){
			Rin = 0;
		}
		Rremain++;
	}
}


//********************************************************************
//	シリアル送信データレジスタ空き割り込み
//********************************************************************
ISR(USART1_UDRE_vect){

	if (Sremain){
		UDR1 = Sbuff[Sout++];
		if (Sout >= sizeof(Sbuff))	Sout = 0;
		--Sremain;
	}
}


//********************************************************************
//	シリアル送信完了割り込み
//	プログラミングモードで使う
//********************************************************************
ISR(USART1_TX_vect){

	MODE_OFF;				// MODEピン開放
	UCSR1B &= ~(1<<TXEN1);	// TX disable
	UCSR1B |= (1<<RXEN1);	// 受信ON(エコーキャンセルOFF)
}


//********************************************************************
//	シリアルポート送信
//********************************************************************
void send_com(unsigned char data){

	if (Sremain || (!(UCSR1A & (1<<UDRE1)))){
		cli();
		Sbuff[Sin++] = data;
		++Sremain;
		sei();
		if (Sin >= sizeof(Sbuff))	Sin = 0;
	} else {
		UCSR1B &= ~(1<<TXCIE1);				// シリアル送信完了割り込み禁止
		UCSR1B |= (1<<TXEN1)|(1<<RXEN1)|(1<<UDRIE1);	// TX, RX enable, 送信データ空き割り込み
		UDR1 = data;
	}
}


//********************************************************************
//	MODEピンへデータ送信
//********************************************************************
void send_mode(unsigned char data){

	UCSR1B &= ~((1<<RXEN1)|(1<<UDRIE1));	// 受信OFF(エコーキャンセルON), 送信データ空き割り込み禁止
	UCSR1B |= (1<<TXEN1)|(1<<TXCIE1);	// TX enable, 送信完了割り込み許可
	MODE_ON;				// MODE出力ON
	UDR1 = data;
}


//********************************************************************
//	ボーレート変更コールバック
//	cdc162.c の usbcdc_setup() から呼び出される
//********************************************************************
void callback_set_baud(long baud){

	Baud = baud;
}


//********************************************************************
//	時間待ち
//	内部で Timer を使うことに注意
//********************************************************************
void wait_timer(unsigned char wait){

	Timer = wait;
	while(Timer){
		WDR;
		SLEEP;
		usbcdc_poll();	// USBポーリング処理
	}
}


//********************************************************************
//	R8Cプログラムモード設定
//********************************************************************
int enter_program(void){
	int		i;
	int		ch = -1;

	LED_PROG_ON;
	UCSR1B &= ~((1<<TXEN1)|(1<<RXEN1));	// TX, RX disable
	PEN2_H;
	PEN1_L;				// buffer ON
	MODE_ON;
	MODE_L;
	RESET_L;			// RESET
	wait_timer(10);		// 100ms
	RESET_H;
	wait_timer(7);		// 70ms
	MODE_H;
	wait_timer(15);		// 150ms

	send_mode(0xB0);	// 0xB0 送信
	for (i = 0; i < 16; i++){
		wait_timer(2);	// 20ms
		send_mode(0);	// 0 送信
	}

	// USBから 0以外のコマンドが入力されるまで待つ
	Timer = 100 - 64;	// 最大１秒
	while((ch < 0)&& Timer){
		usbcdc_poll();
		if (usbcdc_can_getc(1)){
			ch = usbcdc_getc();			// USB受信
			if (ch == 0){
				ch = -1;				// 同期用ゼロ読み捨て
			}

		}
	}

	Rremain = Rin = Rout = 0;	// 念のため受信バッファ初期化

	return ch;
}


//********************************************************************
//	R8Cプログラムモード終了
//********************************************************************
void quit_program(void){

	RESET_L;			// RESET
	wait_timer(10);		// 100ms
	RESET_H;
	PEN1_H;
	PEN2_L;
	LED_PROG_OFF;
	UCSR1B |= (1<<RXEN1);	// 受信ON
}


//********************************************************************
//	メインルーチン
//********************************************************************
int main(void){
	char	target_power = 0;
	char	program_mode = 0;
	int		skip_count = 0;
	int		size_count = 0;
	unsigned char	bit_rate_setting = 0;
	char	bit_rate_set_flag = 0;
	long	baud = 0;
	int		ch = -1;

	cli();

	REGCR &= ~(1<<REGDIS);	// USBレギュレータON

	PORTB = PORTB_INIT;		// ポート初期設定
	DDRB = PORTB_DIR;
	PORTC = PORTC_INIT;
	DDRC = PORTC_DIR;
	PORTD = PORTD_INIT;
	DDRD = PORTD_DIR;
	
	WDR;
	WDTCSR = (1<<WDCE)|(1<<WDE);
	WDTCKD = 0;
	while(WDTCSR & (1<<WDCE));
	WDTCSR = (1<<WDCE)|(1<<WDE);
	WDTCSR = (1<<WDE)|(1<<WDP2)|(1<<WDP1);	// ウォッチドッグ 約2秒

    /* set system_clock pre scaler 1/8 => 1/1 */
    CLKPR = (1<<CLKPCE);
    CLKPR = 0;

	// 100Hzタイマ設定
	TCCR1A = 0;
	TCCR1B = (1<<WGM12)|(1<<CS11)|(1<<CS10);	// CTC  1/64
	TCCR1C = 0;
	OCR1AH = (CPU_CLOCK/64/100 - 1) >> 8;
	OCR1AL = (unsigned char)(CPU_CLOCK/64/100 - 1);
	TIMSK1 = (1<<OCIE1A);		// 割り込み許可

	// USB設定
    usbcdc_init();

	// USART デフォルト 9600bps
	UBRR1H = (CPU_CLOCK/16/9600 - 1) >> 8;
	UBRR1L = (unsigned char)(CPU_CLOCK/16/9600 - 1);
	UCSR1A = 0;
	UCSR1C = (1<<UCSZ11)|(1<<UCSZ10);	// 8bit/stop 1/parity non
	UCSR1D = 0;							// CTS, RTS disable
	//UCSR1B = (1<<TXCIE1)|(1<<RXCIE1);	// 送信完了・受信割り込み許可
	UCSR1B = (1<<RXEN1)|(1<<RXCIE1);	// 受信割り込み許可

    sei();

	if (!TARGET_POWER){
		target_power = 1;
	}
	
	// ************ メインループ **********
    for (;;){

		WDR;
		//SLEEP;	

		usbcdc_poll();		// USBポーリング処理

		// ターゲット電源 ON/OFF処理
		if (target_power == 0){
			if (TARGET_POWER){
				wait_timer(10);		// 100ms時間待ち
				if (TARGET_POWER){	// 電源が入った
					LED_POWER_ON;
					LED_PROG_OFF;
					PEN1_H;
					PEN2_L;
					Rremain = Rin = Rout = 0;
					UCSR1B |= (1<<RXEN1);	// RX enable
					target_power = 1;
				}
			}
		} else if (!TARGET_POWER){	// 電源が切れた
			PEN1_H;
			PEN2_H;
			MODE_OFF;
			UCSR1B &= ~((1<<RXEN1)|(1<<TXEN1));	// RX, TX disable
			Rremain = Rin = Rout = 0;
			LED_POWER_OFF;
			LED_PROG_ON;
			target_power = 0;
			program_mode = 0;
		}

		// ボーレート設定
		if (target_power && Baud){
			// ボーレート設定 115200と57600は誤差が大きいので特別設定
			baud = Baud;
			if (Baud == 115200){
				UBRR1H = 0;
#if (CPU_CLOCK == 16000000UL)
				UBRR1L = 16;	// 2.1%
#else
				UBRR1L = 8;		// -3.5%
#endif
				UCSR1A |= (1<<U2X1);	// 1/8
				UCSR1C |= (1<<USBS1);	// 送信時のみ stop2 受信時 stop1
			} else if(Baud == 57600){
				UBRR1H = 0;
#if (CPU_CLOCK == 16000000UL)
				UBRR1L = 34;	// -0.8%
				UCSR1C &= ~(1<<USBS1);	// 常に stop1
#else
				UBRR1L = 16;	// 2.1%
				UCSR1C |= (1<<USBS1);	// 送信時のみ stop2 受信時 stop1
#endif
				UCSR1A |= (1<<U2X1);	// 1/8
			} else {
				if (program_mode && (bit_rate_set_flag == 2)){
					// R8Cボーレート変更コマンドの設定値より
					UBRR1H = 0;
#if (CPU_CLOCK == 16000000UL)
					UBRR1L = (bit_rate_setting + 1) * 2 - 1;
#else
					UBRR1L = bit_rate_setting;
#endif
				} else {
					UBRR1H = (CPU_CLOCK/16/Baud - 1) >> 8;
					UBRR1L = (unsigned char)(CPU_CLOCK/16/Baud - 1);
				}
				UCSR1A &= ~(1<<U2X1);	// 1/16
				UCSR1C &= ~(1<<USBS1);	// 常に stop1
			}
			bit_rate_set_flag = 0;
			Baud = 0;
		}
		
		// USB受信データ引き受け
		if ((ch < 0)&&(usbcdc_can_getc(1))){

			ch = usbcdc_getc();	// USB受信

			if (!program_mode && !DataIntervalTimer
				&&(baud == 9600)&&(ch == 0xB0)){

				ch = enter_program();	// プログラムモードに入る
				program_mode = 1;
				skip_count = 0;
				size_count = 0;
				bit_rate_set_flag = 0;
				ModeTimer = MODE_TIME;
			}

			if ((ch >= 0)&& program_mode){
				// プログラムモード コマンド解析
				if (size_count){
					if (--size_count == 0){
						skip_count = ch;
					}
				} else if (bit_rate_set_flag == 1){
					bit_rate_setting = (unsigned char)ch;
					bit_rate_set_flag = 2;
				} else if (skip_count == 0){
					switch(ch){
					  case 0xFF:
						skip_count = 2;		break;
					  case 0x41:
						skip_count = 258;	break;
					  case 0x49:
						size_count = 4;		break;
					  case 0x20:
						skip_count = 3;		break;
					  case 0xA7:
						skip_count = 1;		break;
					  case 0xF5:
						skip_count = 11;	break;
					  case 0xB5:
						bit_rate_set_flag = 1;
						skip_count = 1;		break;
					  case 0xB0:
						if (baud != 9600){
							ModeTimer = 0;		// プログラムモード終了
							ch = -1;
						}
					  default:
						break;
					}
				} else {
					--skip_count;
				}
			}
		}

		// プログラムモードタイムアウト
		if (program_mode && !ModeTimer){
			quit_program();		// R8C リセット
			program_mode = 0;

			MODE_OFF;				// MODEピン開放
			UCSR1B |= (1<<RXEN1);	// 受信ON(エコーキャンセルOFF)
		}

		// シリアルへデータ送信
		if (ch >= 0){
			if (program_mode){
				if (!IS_MODE){
					send_mode((unsigned char)ch);	// MODEピン送信
					ModeTimer = MODE_TIME;
					ch = -1;
					DataIntervalTimer = DATA_INTERVAL;
				}
			} else {
				if (Sremain < (sizeof(Sbuff) - 1)){
					send_com((unsigned char)ch);	// シリアル送信
					ch = -1;
					DataIntervalTimer = DATA_INTERVAL;
				}
			}
		}

		// シリアル受信データ → USB送信
		while (Rremain){
			unsigned char	c;

			c = Rbuff[Rout++];
			if (Rout >= sizeof(Rbuff)){
				Rout = 0;
			}
			cli();
			--Rremain;
			sei();

			while(!usbcdc_can_putc(1)){
				usbcdc_poll();		// USBポーリング処理
			}
			usbcdc_putc(c);		// USB送信

			if (program_mode){
				ModeTimer = MODE_TIME;
			}
			DataIntervalTimer = DATA_INTERVAL;
		}
	}

    return 0;
}
