/* Name: usb162.c
 * Project: AVR USB driver for AT90USB
 * Author: Koji Suzuki
 * Creation Date: 2007-12-9
 * Tabsize: 8
 * Copyright: (c) 2007 by Koji Suzuki (suz@nmj.sakura.ne.jp)
 * License: Proprietary, free under certain conditions. See Documentation.
 */

#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include "usbconfig.h"
#include "usb162.h"
#include "desc.h"

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))


#define nop()	__asm__ volatile ("nop" )



#define TYPE_CONT	0
#define TYPE_BULK	2
#define TYPE_INT	3

#define SIZE_8		0

#define ONE_BANK	0
#define DUAL_BANK	1

#define EP_CTRL		0


static uint8_t ep_ctrl_size;
int8_t usb_connected;
int8_t (* usb_user_setup)(uint8_t *, uint8_t *);

void send_zlp () {
	bit_clear(UEINTX, TXINI);  // send ZLP 
	while (bit_is_clear(UEINTX, TXINI))
		;
}

int endpoint_setup(uint8_t ep) {
	char *p;
	uint8_t dir,attr,size,limit;
	uint8_t sz,bank;
	if (ep == EP_CTRL) {
		p = usbDescrDevice;
		size = __LPM(p+7);
		sz = SIZE_8;
		while ((8<<sz) < size )
			sz++;
		UENUM = EP_CTRL;
		if (bit_is_clear(UECONX, EPEN)) {
			bit_set(UECONX, EPEN);
			UECFG0X = (TYPE_CONT<<EPTYPE0) | (0<<EPDIR);
			UECFG1X = (sz<<EPSIZE0) | (ONE_BANK<<EPBK0);
			bit_set(UECFG1X,ALLOC);
		} else {
			bit_set(UERST,EP_CTRL);
			nop();
			bit_clear(UERST,EP_CTRL);
		}
#ifdef EEPROM_DEBUG
		{
			uint8_t buf[4];
			buf[0] = UECFG0X;
			buf[1] = UECFG1X;
			buf[2] = UESTA0X;
			buf[3] = UESTA1X;
			eeprom_write_block(buf,
			    (uint8_t *)(0x50 + (ep << 2)),4);
		}
#endif
		ep_ctrl_size = size;
		return size;
	}
	p = usbDescrConfig;
	limit = __LPM(p+2);
	while (p - usbDescrConfig < limit) {
		if ((__LPM(p+1) == USB_DESCR_ENDPOINT) 
		     && ((__LPM(p+2) & 0xf) == ep))  { // endpoint desc.
			dir = (__LPM(p+2) & 0x80) == 0x80; // 0x80 : IN
			attr = __LPM(p+3); // bulk:2 , interrupt:3
			size = __LPM(p+4);
			sz = SIZE_8;
			while ((8<<sz) < size )
				sz++;
			bank = ONE_BANK;
			if ((ep >= 3) && (attr == 2)) {
#if (USE_DUAL_BANK_BULK_IN)
				if (dir) {
					bank = DUAL_BANK;
				}
#endif
#if (USE_DUAL_BANK_BULK_OUT)
				if (!dir) {
					bank = DUAL_BANK;
				}
#endif
			}
			UENUM = ep;
			if (bit_is_clear(UECONX, EPEN)) {
				bit_set(UECONX, EPEN);
				UECFG0X = (attr<<EPTYPE0) | dir;
				UECFG1X = (sz<<EPSIZE0) | (bank<<EPBK0);
				bit_set(UECFG1X,ALLOC);
			} else {
				bit_set(UERST,ep);
				nop();
				bit_clear(UERST,ep);
			}
#ifdef EEPROM_DEBUG
			{
				uint8_t buf[4];
				buf[0] = UECFG0X;
				buf[1] = UECFG1X;
				buf[2] = UESTA0X;
				buf[3] = UESTA1X;
				eeprom_write_block(buf,
				    (uint8_t *)(0x50 + (ep << 2)),4);
			}
#endif
			return size;
		}
		p += __LPM(p);
	}
	return 0;
}

void start_pll() {
	if (bit_is_clear(PLLCSR,PLOCK)) {
#if F_CPU==8000000UL
		PLLCSR = (0<<PLLP0) | (1<<PLLE);
#else
		PLLCSR = (1<<PLLP0) | (1<<PLLE);
#endif
		while (bit_is_clear(PLLCSR,PLOCK))
			;
	}
	bit_clear(USBCON,FRZCLK); // enable usb clock 
	while (bit_is_clear(UDINT, SOFI))
		;
}

static inline void usbReset() {
#if USB162_SOFT_RESET
	int8_t i;

	usb_connected = 0;
	GPIOR2 = 0;
	bit_set(USBCON,USBE);     // enable usb
	for (i=0; i<= 4; i++) {
		endpoint_setup(i);
	}
#endif
}

void usbInit() {
	int8_t i;

#ifdef EEPROM_DEBUG
    {
	    int16_t i;
	    uint8_t buf[4];
	    memset(buf, 0xff, 4);
	    for (i=0; i< 0x80; i+=4) {
		    eeprom_write_block(buf, (void *)i, 4);
	    }
    }
#endif
	USBCON = (1<<FRZCLK) | (1<<DETACH);
	UDIEN = 0;
	UEIENX = 0;
	bit_set(USBCON,USBE);     // enable usb
	bit_clear(UDCON, DETACH); // connect internal pull-up 
	start_pll();
#if USB162_SOFT_RESET
	bit_clear(UDCON, RSTCPU);
#else
	bit_set(UDCON, RSTCPU);   // reset cpu @ USB_RESET
#endif
	bit_clear(UDADDR, ADDEN); // disable addrss
	for (i=0; i<= 4; i++) {
		endpoint_setup(i);
	}
	//bit_set(UDIEN, SUSPE);    // Enable Suspend int
	bit_set(UDIEN, EORSTE);	  // Enable Reset int
	//bit_set(UDIEN, WAKEUPE);  // Enable Wakeup int (one time)
}

ISR(USB_GEN_vect) {
	if (bit_is_set(UDINT,EORSTI)) {
		// USB Reset
		usbReset();
		bit_set(GPIOR2, USB162_RECV_RESET);
		bit_clear(UDINT,EORSTI);
	}
	if (bit_is_set(UDINT, WAKEUPI)) {
		bit_clear(UDIEN, WAKEUPE);
		start_pll();
		bit_clear(UDINT, WAKEUPI);
	}
}

void usb_reply_setup(uint8_t requested, uint8_t *p, uint8_t len, int8_t mode) {
	uint8_t fifo_len = 0;
	uint8_t written = 0;

	bit_clear(UEINTX, NAKOUTI);
	while (len) {
		if (bit_is_set(UEINTX, NAKOUTI)) { // got ZLP -- abort
			goto ret;
		}
		UEDATX = (mode == 0)?(*p):(__LPM(p));
		p++;
		fifo_len++;
		written++;
		len--;
		if (written >= requested) break;
		if (fifo_len == ep_ctrl_size) {
			bit_clear(UEINTX, TXINI); // send
			while (bit_is_clear(UEINTX, TXINI)
					&& bit_is_clear(UEINTX, NAKOUTI))
				;
			fifo_len = 0;
		}
	}
	if (bit_is_set(UEINTX, NAKOUTI)) { // got ZLP -- abort
		goto ret;
	}
ret:
	if (fifo_len || (written != requested)) { // send ZLP
		bit_clear(UEINTX, TXINI);
	}
	// wait ZLP (from HOST)
	while (bit_is_clear(UEINTX, RXOUTI)
		&& bit_is_clear(UEINTX, NAKOUTI))
		;
	bit_clear(UEINTX, NAKOUTI);
	bit_clear(UEINTX, RXOUTI);
}

uint8_t reply_data[8];
uint8_t setup_data[8];
uint8_t usb_config_value;
#ifdef EEPROM_DEBUG
uint16_t seq_no;
#endif

void usb_process_setup(){
	uint8_t reply_requested;
	uint8_t request;
	uint8_t request_type;
	uint8_t rcpt;
	int8_t r;

	bit_clear(UEINTX,RXOUTI);
	for (r=0; r<8; r++) {
		setup_data[r] = UEDATX;
	}
	request_type = setup_data[USB_RQ_TYPE];
	request = setup_data[USB_RQ_REQ];
#ifdef EEPROM_DEBUG
	if (seq_no < 0x60) {
		uint8_t buf[4];
		buf[0] = request_type;
		buf[1] = request;
		buf[2] = setup_data[USB_RQ_VALUE_L];
		buf[3] = setup_data[USB_RQ_VALUE_H];
#ifdef EEPROM_DEBUG_LENGTH
		buf[0] = setup_data[USB_RQ_LENGTH_L];
#endif
		eeprom_write_block(buf, (uint8_t *)seq_no, 4);
		seq_no +=4;
	}
#endif
	bit_clear(UEINTX,RXSTPI);

	if (setup_data[USB_RQ_LENGTH_H]) {
		reply_requested = 254;
	} else {
		reply_requested = setup_data[USB_RQ_LENGTH_L];
	}
	if ((request_type & USB_RQ_TYPE_MASK) == USB_RQ_TYPE_STANDARD) {
		switch (request) {
		  case USB_RQ_GET_STATUS:
			rcpt = request_type & USB_RQ_RCPT_MASK;
			reply_data[0] = 0;
			if (rcpt == USB_RQ_RCPT_DEVICE) {
				reply_data[0] = (__LPM(usbDescrConfig+7) & USB_ATTR_SELFPOWER)?1:0;
			}
			if (rcpt == USB_RQ_RCPT_ENDPOINT 
				&& setup_data[USB_RQ_INDEX_L] == 0x81)
				reply_data[0] = 1;
			reply_data[1] = 0;
			usb_reply_setup(reply_requested, reply_data, 2, 0);
			break;
		  case USB_RQ_SET_ADDRESS:
#if 0
			if (!setup_data[USB_RQ_VALUE_L] && !UDADDR) {
				bit_set(UECONX, STALLRQ);
				return;
			}
#endif
			UDADDR = setup_data[USB_RQ_VALUE_L] & 0x7f;
			send_zlp();

			bit_set(UDADDR, ADDEN); // enable addrss
			break;
		  case USB_RQ_GET_DESCRIPTOR:
			if ( setup_data[USB_RQ_VALUE_H] == 1) {
				usb_reply_setup(reply_requested,
								(uint8_t *)usbDescrDevice,
								__LPM(usbDescrDevice), 1);
			} else if (setup_data[USB_RQ_VALUE_H] == 2) {
				usb_reply_setup(reply_requested,
								(uint8_t *)usbDescrConfig,
								__LPM(usbDescrConfig+2), 1);
			} else if (setup_data[USB_RQ_VALUE_H] == 3) {
			    if (setup_data[USB_RQ_VALUE_L] == 0) {
					usb_reply_setup(reply_requested,
									(uint8_t *)usbDescrString0,
									__LPM(usbDescrString0), 1);
			    } else if (setup_data[USB_RQ_VALUE_L] == 1) {
					usb_reply_setup(reply_requested,
									(uint8_t *)usbDescrString1,
									__LPM(usbDescrString1), 1);
			    } else if (setup_data[USB_RQ_VALUE_L] == 2) {
					usb_reply_setup(reply_requested,
									(uint8_t *)usbDescrString2,
									__LPM(usbDescrString2), 1);
			    } else if (setup_data[USB_RQ_VALUE_L] == 3) {
					usb_reply_setup(reply_requested,
									(uint8_t *)usbDescrString3,
									__LPM(usbDescrString3), 1);
			    }
#if 0
			} else if (setup_data[USB_RQ_VALUE_H] == USB_DESCR_HID_REPORT) {
#endif
			} else {
				bit_set(UECONX, STALLRQ);
			}
			break;
		  case USB_RQ_GET_CONFIGURATION:
			reply_data[0] = usb_config_value;
			usb_reply_setup(reply_requested, reply_data, 1, 0);
			break;
		  case USB_RQ_SET_CONFIGURATION:
			usb_config_value = setup_data[USB_RQ_VALUE_L];
			bit_clear(UEINTX, TXINI);  // send ZLP ?
			usb_connected = 1;
			break;
		  case USB_RQ_GET_INTERFACE:
			reply_data[0] = 0;
			usb_reply_setup(reply_requested, reply_data, 1, 0);
			break;
		  case USB_RQ_CLEAR_FEATURE:
		  case USB_RQ_SET_FEATURE:
			if (setup_data[USB_RQ_VALUE_L] == 0 &&
				setup_data[USB_RQ_INDEX_L] == 0x81) {
				/* EP1 HALT */
				bit_set(UECONX, STALLRQ);
			}
			bit_clear(UEINTX, TXINI); // send ZLP ?
			break;
		  case USB_RQ_SET_INTERFACE:
			bit_clear(UEINTX, TXINI); // send ZLP ?
			break;
		  default:
			bit_set(UECONX, STALLRQ);
			break;
		}
	} else if (usb_user_setup) {
		r = (usb_user_setup)(setup_data,reply_data);
		if (r >= 0) usb_reply_setup(reply_requested, reply_data, r, 0);
		else if (r == -1) send_zlp();
		else bit_set(UECONX, STALLRQ);
	} else {
		bit_set(UECONX, STALLRQ);
	}
}

void usbPoll(){
	if (bit_is_set(GPIOR2, USB162_RECV_RESET)) {
#ifdef EEPROM_DEBUG
		if (seq_no < 0x60) {
			uint8_t buf[4];
			buf[0] = 0xaa;
			buf[1] = 0xaa;
			buf[2] = 0xaa;
			buf[3] = 0xaa;
			eeprom_write_block(buf, (uint8_t *)seq_no, 4);
			seq_no +=4;
		}
#endif
		bit_clear(GPIOR2, USB162_RECV_RESET);
	}
	bit_clear(UDINT, SOFI);
	UENUM = EP_CTRL;
	if (bit_is_set(UEINTX, RXSTPI)) { // recieve SETUP 
		usb_process_setup();
	}

	if (bit_is_set(UEINTX,NAKINI)) {
		bit_clear(UEINTX,NAKINI);
	}
}
